@@ -3,8 +3,12 @@ from enum import auto
 from typing import List
 
 from jsonrpcclient.clients.http_client import HTTPClient
+from jsonrpcclient.requests import Request
 
 import logging
+
+from jsonrpcclient.exceptions import ReceivedNon2xxResponseError
+
 from common.cs_enum import AutoName
 
 DEFAULT_LEDGER_BOOK_ID = "1"
@@ -14,6 +18,7 @@ log = logging.getLogger(__name__)
 
 class OperationType(AutoName):
     TRANSFER = auto()
+    CUSTOM = auto()
     VOID = auto()
 
 
@@ -30,16 +35,44 @@ class LedgerClient:
         if ledger_read_only_endpoint is not None:
             self.read_only_rpc_client = HTTPClient(ledger_read_only_endpoint)
 
-    def _execute(self, method: str, *args, read_only_method: bool = False):
+    def _execute(self, method: str, *args, read_only_method: bool = False, retry_index=0):
         start_time = int(time.time()*1000)
-        if read_only_method is True and getattr(self, "read_only_rpc_client", None) is not None:
-            rpc_response = self.read_only_rpc_client.request(method, *args)
-        else:
-            rpc_response = self.rpc_client.request(method, *args)
+        try:
+            if read_only_method is True and getattr(self, "read_only_rpc_client", None) is not None:
+                rpc_response = self.read_only_rpc_client.request(method, *args)
+            else:
+                rpc_response = self.rpc_client.request(method, *args)
+        except ReceivedNon2xxResponseError as e:
+            if e.code == 502 and retry_index <= 3:
+                log.warning(f"Retrying 502 response", exc_info=1)
+                return self._execute(method, *args, read_only_method=read_only_method, retry_index=retry_index+1)
+            raise e
         end_time = int(time.time()*1000)
         log.info(f"LedgerClient: {method}: time_taken: {end_time-start_time}ms")
         return rpc_response.data.result
 
+    def _bulk_execute(self,request_list: List, read_only_method: bool = False, retry_index=0):
+        start_time = int(time.time()*1000)
+        try:
+            if read_only_method is True and getattr(self, "read_only_rpc_client", None) is not None:
+                rpc_response = self.read_only_rpc_client.send(request_list)
+            else:
+                rpc_response = self.rpc_client.send(request_list)
+        except ReceivedNon2xxResponseError as e:
+            if e.code == 502 and retry_index <= 3:
+                log.warning(f"Retrying 502 response", exc_info=1)
+                return self._bulk_execute(request_list, read_only_method=read_only_method, retry_index=retry_index+1)
+            raise e
+        end_time = int(time.time()*1000)
+        log.info(f"LedgerClient: time_taken: {end_time-start_time}ms")
+        result = []
+        res_id_map = {}
+        for each_response in rpc_response.data:
+            res_id_map[each_response.id] = each_response.result
+        for each_request in request_list:
+            result.append(res_id_map[each_request['id']])
+        return result
+
     def create_book(self, name: str, min_balance: str = None, metadata: dict = None):
         book_info = dict()
         book_info["name"] = name
@@ -66,6 +99,19 @@ class LedgerClient:
         filter = {"toTime": timestamp} if timestamp else None
         return self._execute("getBalances", book_id, asset_id, metadata_filter, filter, read_only_method=True)
 
+    def get_book_balances_bulk(self, params_list: list = []):
+        request_list = []
+        for params in params_list:
+            if not params.get('book_id'):
+                raise Exception('invalid input to get_book_balances_bulk: book_id missing')
+            params['asset_id'] = params.get('asset_id',None)
+            params['metadata_filter'] = params.get('metadata_filter', None)
+            params['filter'] = {"toTime": params['timestamp']} if params.get('timestamp', None) else None
+            request_list.append(
+                Request("getBalances", params['book_id'], params['asset_id'], params['metadata_filter'], params['filter'])
+            )
+        return self._bulk_execute(request_list, read_only_method = True)
+
     def get_operation(self, operation_id: str):
         return self._execute("getOperation", operation_id, read_only_method=True)
 
@@ -96,6 +142,9 @@ class LedgerClient:
             raise Exception(result["rejectionReason"])
         return result
 
+    def get_diff(self, from_time: int, to_time: int):
+        return self._execute("getDiff", from_time, to_time, read_only_method=True)
+
     def post_void(self, operation_id_list: list, memo: str, metadata: dict = None):
         void = dict()
         void["operationIds"] = operation_id_list
